@@ -23,6 +23,7 @@ use ibc_proto::ibc::core::channel::v1::QueryConnectionChannelsRequest;
 
 use crate::chain::counterparty::{channel_connection_client, channel_state_on_destination};
 use crate::chain::handle::ChainHandle;
+use crate::chain::tx::TrackedMsgs;
 use crate::channel::version::ResolveContext;
 use crate::connection::Connection;
 use crate::foreign_client::{ForeignClient, HasExpiredOrFrozenError};
@@ -739,9 +740,11 @@ impl<ChainA: ChainHandle, ChainB: ChainHandle> Channel<ChainA, ChainB> {
     pub fn build_chan_open_init_and_send(&self) -> Result<IbcEvent, ChannelError> {
         let dst_msgs = self.build_chan_open_init()?;
 
+        let tm = TrackedMsgs::new(dst_msgs, "ChannelOpenInit");
+
         let events = self
             .dst_chain()
-            .send_messages_and_wait_commit(dst_msgs)
+            .send_messages_and_wait_commit(tm)
             .map_err(|e| ChannelError::submit(self.dst_chain().id(), e))?;
 
         // Find the relevant event for channel open init
@@ -903,9 +906,11 @@ impl<ChainA: ChainHandle, ChainB: ChainHandle> Channel<ChainA, ChainB> {
     pub fn build_chan_open_try_and_send(&self) -> Result<IbcEvent, ChannelError> {
         let dst_msgs = self.build_chan_open_try()?;
 
+        let tm = TrackedMsgs::new(dst_msgs, "ChannelOpenTry");
+
         let events = self
             .dst_chain()
-            .send_messages_and_wait_commit(dst_msgs)
+            .send_messages_and_wait_commit(tm)
             .map_err(|e| ChannelError::submit(self.dst_chain().id(), e))?;
 
         // Find the relevant event for channel open try
@@ -988,9 +993,11 @@ impl<ChainA: ChainHandle, ChainB: ChainHandle> Channel<ChainA, ChainB> {
         ) -> Result<IbcEvent, ChannelError> {
             let dst_msgs = channel.build_chan_open_ack()?;
 
+            let tm = TrackedMsgs::new(dst_msgs, "ChannelOpenAck");
+
             let events = channel
                 .dst_chain()
-                .send_messages_and_wait_commit(dst_msgs)
+                .send_messages_and_wait_commit(tm)
                 .map_err(|e| ChannelError::submit(channel.dst_chain().id(), e))?;
 
             // Find the relevant event for channel open ack
@@ -1084,9 +1091,10 @@ impl<ChainA: ChainHandle, ChainB: ChainHandle> Channel<ChainA, ChainB> {
         ) -> Result<IbcEvent, ChannelError> {
             let dst_msgs = channel.build_chan_open_confirm()?;
 
+            let tm = TrackedMsgs::new(dst_msgs, "ChannelOpenConfirm");
             let events = channel
                 .dst_chain()
-                .send_messages_and_wait_commit(dst_msgs)
+                .send_messages_and_wait_commit(tm)
                 .map_err(|e| ChannelError::submit(channel.dst_chain().id(), e))?;
 
             // Find the relevant event for channel open confirm
@@ -1147,9 +1155,11 @@ impl<ChainA: ChainHandle, ChainB: ChainHandle> Channel<ChainA, ChainB> {
     pub fn build_chan_close_init_and_send(&self) -> Result<IbcEvent, ChannelError> {
         let dst_msgs = self.build_chan_close_init()?;
 
+        let tm = TrackedMsgs::new(dst_msgs, "ChannelCloseInit");
+
         let events = self
             .dst_chain()
-            .send_messages_and_wait_commit(dst_msgs)
+            .send_messages_and_wait_commit(tm)
             .map_err(|e| ChannelError::submit(self.dst_chain().id(), e))?;
 
         // Find the relevant event for channel close init
@@ -1226,9 +1236,11 @@ impl<ChainA: ChainHandle, ChainB: ChainHandle> Channel<ChainA, ChainB> {
     pub fn build_chan_close_confirm_and_send(&self) -> Result<IbcEvent, ChannelError> {
         let dst_msgs = self.build_chan_close_confirm()?;
 
+        let tm = TrackedMsgs::new(dst_msgs, "ChannelCloseConfirm");
+
         let events = self
             .dst_chain()
-            .send_messages_and_wait_commit(dst_msgs)
+            .send_messages_and_wait_commit(tm)
             .map_err(|e| ChannelError::submit(self.dst_chain().id(), e))?;
 
         // Find the relevant event for channel close confirm
