@@ -6,7 +6,7 @@ use ibc::core::ics24_host::identifier::{ChainId, ChannelId, ClientId, PortChanne
 use ibc::events::IbcEvent;
 
 use crate::error::Error;
-use crate::foreign_client::ForeignClientError;
+use crate::foreign_client::{ForeignClientError, HasExpiredOrFrozenError};
 use crate::supervisor::Error as SupervisorError;
 
 define_error! {
@@ -74,6 +74,11 @@ define_error! {
             [ Error ]
             |e| { format_args!("failed during a query to chain id {0}", e.chain_id) },
 
+        QueryAppVersion
+            { chain_id: ChainId }
+            [ Error ]
+            |e| { format_args!("failed during a query for the app version to chain id {0}", e.chain_id) },
+
         QueryChannel
             { channel_id: ChannelId }
             [ SupervisorError ]
@@ -191,5 +196,28 @@ define_error! {
                 format_args!("channel object cannot be built from event: {}",
                     e.event)
             },
+
+        InvalidPortId
+            { port_id: PortId }
+            | e | {
+                format_args!("could not resolve channel version because the port is invalid: {0}",
+                    e.port_id)
+            },
+
+    }
+}
+
+impl HasExpiredOrFrozenError for ChannelErrorDetail {
+    fn is_expired_or_frozen_error(&self) -> bool {
+        match self {
+            Self::ClientOperation(e) => e.source.is_expired_or_frozen_error(),
+            _ => false,
+        }
+    }
+}
+
+impl HasExpiredOrFrozenError for ChannelError {
+    fn is_expired_or_frozen_error(&self) -> bool {
+        self.detail().is_expired_or_frozen_error()
     }
 }
