@@ -58,7 +58,19 @@ pub fn get_all_events(
                 if *query == queries::ibc_channel().to_string() {
                     if let Some(mut chan_event) = ChannelEvents::try_from_tx(abci_event) {
                         chan_event.set_height(height);
-                        tracing::trace!("extracted ibc_channel event {:?}", chan_event);
+                        let _span = tracing::trace_span!("ibc_channel event").entered();
+                        tracing::trace!("extracted {:?}", chan_event);
+                        if matches!(chan_event, IbcEvent::SendPacket(_)) {
+                            // Should be the same as the hash of tx_result.tx?
+                            if let Some(hash) = result
+                                .events
+                                .as_ref()
+                                .and_then(|events| events.get("tx.hash"))
+                                .and_then(|values| values.get(0))
+                            {
+                                tracing::trace!(event = "SendPacket", "tx hash: {}", hash);
+                            }
+                        }
                         vals.push((height, chan_event));
                     }
                 }
