@@ -6,6 +6,7 @@ use itertools::Itertools;
 use prost_types::Any;
 use tracing::{debug, error, info, trace, warn};
 
+use crate::chain::tx::TrackedMsgs;
 use crate::error::Error as RelayerError;
 use flex_error::define_error;
 use ibc::core::ics02_client::client_consensus::{
@@ -29,6 +30,7 @@ use ibc::timestamp::{Timestamp, TimestampOverflowError};
 use ibc::tx_msg::Msg;
 use ibc::Height;
 use ibc_proto::ibc::core::client::v1::QueryConsensusStatesRequest;
+use tendermint_light_client_verifier::types::TrustThreshold;
 
 use crate::chain::handle::ChainHandle;
 
@@ -138,10 +140,22 @@ define_error! {
             }
             [ RelayerError ]
             |e| {
-                format_args!("failed while querying Tx for client {0} on chain id: {1}",
+                format_args!("failed while querying for client {0} on chain id {1}",
                     e.client_id, e.chain_id)
             },
 
+        ClientConsensusQuery
+            {
+                client_id: ClientId,
+                chain_id: ChainId,
+                height: Height
+            }
+            [ RelayerError ]
+            |e| {
+                format_args!("failed while querying for client consensus state {0} on chain id {1} for height {2}",
+                    e.client_id, e.chain_id, e.height)
+            },
+
         ClientUpgrade
             {
                 client_id: ClientId,
@@ -269,6 +283,16 @@ pub struct ForeignClient<DstChain: ChainHandle, SrcChain: ChainHandle> {
     pub src_chain: SrcChain,
 }
 
+/// Optional onfiguration parameters for the
+/// CreateClient command. If set the options override the defaults
+/// taken from the configuration of the destination chain.
+#[derive(Debug, Default)]
+pub struct CreateParams {
+    pub clock_drift: Option<Duration>,
+    pub trusting_period: Option<Duration>,
+    pub trust_threshold: Option<TrustThreshold>,
+}
+
 /// Used in Output messages.
 /// Provides a concise description of a [`ForeignClient`],
 /// using the format:
@@ -433,9 +457,11 @@ impl<DstChain: ChainHandle, SrcChain: ChainHandle> ForeignClient<DstChain, SrcCh
 
         msgs.push(msg_upgrade);
 
+        let tm = TrackedMsgs::new(msgs, "upgrade client");
+
         let res = self
             .dst_chain
-            .send_messages_and_wait_commit(msgs)
+            .send_messages_and_wait_commit(tm)
             .map_err(|e| {
                 ForeignClientError::client_upgrade(
                     self.id.clone(),
@@ -463,7 +489,10 @@ impl<DstChain: ChainHandle, SrcChain: ChainHandle> ForeignClient<DstChain, SrcCh
     }
 
     /// Lower-level interface for preparing a message to create a client.
-    pub fn build_create_client(&self) -> Result<MsgCreateAnyClient, ForeignClientError> {
+    pub fn build_create_client(
+        &self,
+        params: &CreateParams,
+    ) -> Result<MsgCreateAnyClient, ForeignClientError> {
         // Get signer
         let signer = self.dst_chain.get_signer().map_err(|e| {
             ForeignClientError::client_create(
@@ -485,21 +514,29 @@ impl<DstChain: ChainHandle, SrcChain: ChainHandle> ForeignClient<DstChain, SrcCh
             )
         })?;
 
+        let mut config = self.dst_chain.config().map_err(|e| {
+            ForeignClientError::client_create(
+                self.src_chain.id(),
+                format!(
+                    "failed while querying dst chain ({}) for its configuration",
+                    self.dst_chain.id()
+                ),
+                e,
+            )
+        })?;
+        if let Some(d) = params.clock_drift {
+            config.clock_drift = d;
+        }
+        if let Some(p) = params.trusting_period {
+            config.trusting_period = Some(p);
+        }
+        if let Some(t) = params.trust_threshold {
+            config.trust_threshold = t;
+        }
+
         let client_state = self
             .src_chain
-            .build_client_state(
-                latest_height,
-                self.dst_chain.config().map_err(|e| {
-                    ForeignClientError::client_create(
-                        self.src_chain.id(),
-                        format!(
-                            "failed while querying dst chain ({}) for its configuration",
-                            self.dst_chain.id()
-                        ),
-                        e,
-                    )
-                })?,
-            )
+            .build_client_state(latest_height, config)
             .map_err(|e| {
                 ForeignClientError::client_create(
                     self.src_chain.id(),
@@ -533,12 +570,18 @@ impl<DstChain: ChainHandle, SrcChain: ChainHandle> ForeignClient<DstChain, SrcCh
     }
 
     /// Returns the identifier of the newly created client.
-    pub fn build_create_client_and_send(&self) -> Result<IbcEvent, ForeignClientError> {
-        let new_msg = self.build_create_client()?;
+    pub fn build_create_client_and_send(
+        &self,
+        params: &CreateParams,
+    ) -> Result<IbcEvent, ForeignClientError> {
+        let new_msg = self.build_create_client(params)?;
 
         let res = self
             .dst_chain
-            .send_messages_and_wait_commit(vec![new_msg.to_any()])
+            .send_messages_and_wait_commit(TrackedMsgs::new_single(
+                new_msg.to_any(),
+                "create client",
+            ))
             .map_err(|e| {
                 ForeignClientError::client_create(
                     self.dst_chain.id(),
@@ -553,10 +596,12 @@ impl<DstChain: ChainHandle, SrcChain: ChainHandle> ForeignClient<DstChain, SrcCh
 
     /// Sends the client creation transaction & subsequently sets the id of this ForeignClient
     fn create(&mut self) -> Result<(), ForeignClientError> {
-        let event = self.build_create_client_and_send().map_err(|e| {
-            error!("[{}]  failed CreateClient: {}", self, e);
-            e
-        })?;
+        let event = self
+            .build_create_client_and_send(&CreateParams::default())
+            .map_err(|e| {
+                error!("[{}]  failed CreateClient: {}", self, e);
+                e
+            })?;
 
         self.id = extract_client_id(&event)?.clone();
         info!("🍭 [{}]  => {:#?}\n", self, event);
@@ -922,9 +967,11 @@ impl<DstChain: ChainHandle, SrcChain: ChainHandle> ForeignClient<DstChain, SrcCh
             ));
         }
 
+        let tm = TrackedMsgs::new(new_msgs, "update client");
+
         let events = self
             .dst_chain()
-            .send_messages_and_wait_commit(new_msgs)
+            .send_messages_and_wait_commit(tm)
             .map_err(|e| {
                 ForeignClientError::client_update(
                     self.dst_chain.id(),
@@ -1035,7 +1082,12 @@ impl<DstChain: ChainHandle, SrcChain: ChainHandle> ForeignClient<DstChain, SrcCh
             .dst_chain
             .query_consensus_state(self.id.clone(), height, Height::zero())
             .map_err(|e| {
-                ForeignClientError::client_query(self.id.clone(), self.dst_chain.id(), e)
+                ForeignClientError::client_consensus_query(
+                    self.id.clone(),
+                    self.dst_chain.id(),
+                    height,
+                    e,
+                )
             })?;
 
         Ok(res)
@@ -1252,9 +1304,11 @@ impl<DstChain: ChainHandle, SrcChain: ChainHandle> ForeignClient<DstChain, SrcCh
             .to_any(),
         );
 
+        let tm = TrackedMsgs::new(msgs, "evidence");
+
         let events = self
             .dst_chain()
-            .send_messages_and_wait_commit(msgs)
+            .send_messages_and_wait_commit(tm)
             .map_err(|e| {
                 ForeignClientError::misbehaviour(
                     format!(
@@ -1391,7 +1445,7 @@ mod test {
         let b_client = ForeignClient::restore(ClientId::default(), b_chain, a_chain);
 
         // Create the client on chain a
-        let res = a_client.build_create_client_and_send();
+        let res = a_client.build_create_client_and_send(&Default::default());
         assert!(
             res.is_ok(),
             "build_create_client_and_send failed (chain a) with error {:?}",
@@ -1400,7 +1454,7 @@ mod test {
         assert!(matches!(res.unwrap(), IbcEvent::CreateClient(_)));
 
         // Create the client on chain b
-        let res = b_client.build_create_client_and_send();
+        let res = b_client.build_create_client_and_send(&Default::default());
         assert!(
             res.is_ok(),
             "build_create_client_and_send failed (chain b) with error {:?}",
