@@ -14,9 +14,8 @@ pub enum ConfigUpdate {
 
 #[derive(Clone, Debug)]
 pub enum SupervisorCmd {
-    UpdateConfig(ConfigUpdate),
+    UpdateConfig(Box<ConfigUpdate>),
     DumpState(Sender<SupervisorState>),
-    Stop(Sender<()>),
 }
 
 #[derive(Copy, Clone, Debug, PartialEq, Eq)]
