@@ -2,11 +2,11 @@ use core::time::Duration;
 use crossbeam_channel::Receiver;
 use ibc::Height;
 use std::sync::{Arc, Mutex};
-use tracing::{error, trace};
+use tracing::{error, error_span, trace};
 
 use crate::chain::handle::ChainHandle;
 use crate::foreign_client::HasExpiredOrFrozenError;
-use crate::link::{error::LinkError, Link, RelaySummary};
+use crate::link::{error::LinkError, Link};
 use crate::object::Packet;
 use crate::telemetry;
 use crate::util::retry::{retry_with_index, RetryResult};
@@ -47,31 +47,39 @@ pub fn spawn_packet_worker<ChainA: ChainHandle, ChainB: ChainHandle>(
     // Mutex is used to prevent race condition between the packet workers
     link: Arc<Mutex<Link<ChainA, ChainB>>>,
 ) -> TaskHandle {
-    spawn_background_task(
-        format!("PacketWorker({})", link.lock().unwrap().a_to_b),
-        Some(Duration::from_millis(1000)),
-        move || {
-            let relay_path = &link.lock().unwrap().a_to_b;
+    let span = {
+        let relay_path = &link.lock().unwrap().a_to_b;
+        error_span!(
+            "packet",
+            src_chain = %relay_path.src_chain().id(),
+            src_port = %relay_path.src_port_id(),
+            src_channel = %relay_path.src_channel_id(),
+            dst_chain = %relay_path.dst_chain().id(),
+        )
+    };
 
-            relay_path
-                .refresh_schedule()
-                .map_err(handle_link_error_in_task)?;
+    spawn_background_task(span, Some(Duration::from_millis(1000)), move || {
+        let relay_path = &link.lock().unwrap().a_to_b;
 
-            relay_path
-                .execute_schedule()
-                .map_err(handle_link_error_in_task)?;
+        relay_path
+            .refresh_schedule()
+            .map_err(handle_link_error_in_task)?;
 
-            let summary = relay_path.process_pending_txs();
+        relay_path
+            .execute_schedule()
+            .map_err(handle_link_error_in_task)?;
 
-            if !summary.is_empty() {
-                trace!("Packet worker produced relay summary: {:?}", summary);
-            }
+        let summary = relay_path.process_pending_txs();
 
-            telemetry!(packet_metrics(&path, &summary));
+        if !summary.is_empty() {
+            trace!("Packet worker produced relay summary: {:?}", summary);
+        }
 
-            Ok(Next::Continue)
-        },
-    )
+        telemetry!(packet_metrics(&path, &summary));
+        let _path = &path; // avoid unused variable warning when telemetry is disabled
+
+        Ok(Next::Continue)
+    })
 }
 
 pub fn spawn_packet_cmd_worker<ChainA: ChainHandle, ChainB: ChainHandle>(
@@ -83,28 +91,34 @@ pub fn spawn_packet_cmd_worker<ChainA: ChainHandle, ChainB: ChainHandle>(
     path: Packet,
 ) -> TaskHandle {
     let mut is_first_run: bool = true;
-    spawn_background_task(
-        format!("PacketCmdWorker({})", link.lock().unwrap().a_to_b),
-        Some(Duration::from_millis(200)),
-        move || {
-            if let Ok(cmd) = cmd_rx.try_recv() {
-                retry_with_index(retry_strategy::worker_stubborn_strategy(), |index| {
-                    handle_packet_cmd(
-                        &mut is_first_run,
-                        &link.lock().unwrap(),
-                        clear_on_start,
-                        clear_interval,
-                        &path,
-                        cmd.clone(),
-                        index,
-                    )
-                })
-                .map_err(|e| TaskError::Fatal(RunError::retry(e)))?;
-            }
+    let span = {
+        let relay_path = &link.lock().unwrap().a_to_b;
+        error_span!(
+            "packet_cmd",
+            src_chain = %relay_path.src_chain().id(),
+            src_port = %relay_path.src_port_id(),
+            src_channel = %relay_path.src_channel_id(),
+            dst_chain = %relay_path.dst_chain().id(),
+        )
+    };
+    spawn_background_task(span, Some(Duration::from_millis(200)), move || {
+        if let Ok(cmd) = cmd_rx.try_recv() {
+            retry_with_index(retry_strategy::worker_stubborn_strategy(), |index| {
+                handle_packet_cmd(
+                    &mut is_first_run,
+                    &link.lock().unwrap(),
+                    clear_on_start,
+                    clear_interval,
+                    &path,
+                    cmd.clone(),
+                    index,
+                )
+            })
+            .map_err(|e| TaskError::Fatal(RunError::retry(e)))?;
+        }
 
-            Ok(Next::Continue)
-        },
-    )
+        Ok(Next::Continue)
+    })
 }
 
 /// Receives worker commands, which may be:
@@ -124,7 +138,7 @@ fn handle_packet_cmd<ChainA: ChainHandle, ChainB: ChainHandle>(
     cmd: WorkerCmd,
     index: u64,
 ) -> RetryResult<(), u64> {
-    trace!("[{}] handling packet worker command {:?}", link.a_to_b, cmd);
+    trace!("handling command {:?}", cmd);
     let result = match cmd {
         WorkerCmd::IbcEvents { batch } => link.a_to_b.update_schedule(batch),
 
@@ -150,8 +164,8 @@ fn handle_packet_cmd<ChainA: ChainHandle, ChainB: ChainHandle>(
         error!(
             path = %path.short_name(),
             retry_index = %index,
-            "[{}] worker will retry: handling command encountered error: {}",
-            link.a_to_b, e
+            "will retry: handling command encountered error: {}",
+            e
         );
 
         return RetryResult::Retry(index);
@@ -175,16 +189,13 @@ fn handle_packet_cmd<ChainA: ChainHandle, ChainB: ChainHandle>(
 
     if let Err(e) = schedule_result {
         if e.is_expired_or_frozen_error() {
-            error!(
-                "[{}] worker aborting due to expired or frozen client",
-                link.a_to_b
-            );
+            error!("aborting due to expired or frozen client");
             return RetryResult::Err(index);
         } else {
             error!(
                 retry_index = %index,
-                "[{}] worker will retry: schedule execution encountered error: {}",
-                link.a_to_b, e
+                "will retry: schedule execution encountered error: {}",
+                e,
             );
             return RetryResult::Retry(index);
         }
@@ -193,7 +204,7 @@ fn handle_packet_cmd<ChainA: ChainHandle, ChainB: ChainHandle>(
     let summary = link.a_to_b.process_pending_txs();
 
     if !summary.is_empty() {
-        trace!("Packet worker produced relay summary: {:?}", summary);
+        trace!("produced relay summary: {:?}", summary);
     }
 
     telemetry!(packet_metrics(path, &summary));
@@ -201,6 +212,9 @@ fn handle_packet_cmd<ChainA: ChainHandle, ChainB: ChainHandle>(
     RetryResult::Ok(())
 }
 
+#[cfg(feature = "telemetry")]
+use crate::link::RelaySummary;
+
 #[cfg(feature = "telemetry")]
 fn packet_metrics(path: &Packet, summary: &RelaySummary) {
     receive_packet_metrics(path, summary);
